#include "plugin_host_manager.h"

#include "taskflow/taskflow.hpp"

#include "plugin_host.h"
#include "audio/device/audio_device_manager.h"
#include "audio/mixer/channel_interface.h"
#include "audio/mixer/mixer.h"
#include "audio/mixer/mixer_track.h"
#include "misc/singleton/singleton_manager.h"
#include "thread_message/thread_message_hubs.h"
#include "vst2/vst2_plugin_host.h"

void plugin_host_manager::init(singleton_initliazer& initliazer) {
    singleton_t<plugin_host_manager>::init(initliazer);
    auto* mixer_ptr = initliazer.require<mixer>();
    mixer_ptr->on_remove_track.add_raw(this, &plugin_host_manager::on_mixer_track_removed);
}

void plugin_host_manager::release(singleton_release_guard& release_guard) {
    singleton_t<plugin_host_manager>::release(release_guard);
    release_guard.require_release<audio_device_manager>();

    for (const plugin_host* host: plugin_hosts_) {
        delete host;
    }
}

plugin_host* plugin_host_manager::create_instrument_plugin_host(const char* path) {
    auto host = load_plugin(path);
    if (host) {
        register_instrument_plugin(host);
    }
    return host;
}

void plugin_host_manager::remove_instrument_plugin_host(plugin_host* host) {
    g_audio_thread_hub.push_message([host, this]() {
        auto find = std::ranges::find(instrument_plugins_, host);
        if (find != instrument_plugins_.end()) {
            g_mixer.remove_track(host->owner_tracks[0]);
            plugin_hosts_.erase(std::ranges::find(plugin_hosts_, host));
            instrument_plugins_.erase(find);
            g_main_thread_hub.push_message([host, this]() {
                on_instrument_removed.broadcast(host);
                delete host;
            });
        }
    });
}

plugin_host* plugin_host_manager::load_plugin(const char* path) {
    auto host = new vst2_plugin_host();
    try {
        host->load_plugin(path);
    } catch (std::exception& e) {
        spdlog::error("Failed to load plugin: {}", e.what());
        delete host;
        return nullptr;
    }
    host->init_channel_interface();
    host->on_update_buffer_size(g_audio_device_manager.get_buffer_size());
    host->on_update_sample_rate(g_audio_device_manager.get_sample_rate());
    plugin_hosts_.push_back(host);
    return host;
}

void plugin_host_manager::register_instrument_plugin(plugin_host* host) {
    instrument_plugins_.push_back(host);
    g_audio_thread_hub.push_message([this, host]() {
        instrument_track* instrument_track = g_mixer.create_instrument_track(host);
        host->channel->set_input_channel(instrument_track->get_channel_interface()->input_channel_nodes);
        host->channel->set_output_channel(instrument_track->get_channel_interface()->output_channel_nodes);
        host->owner_tracks.push_back(instrument_track);
        host->update_channel_node_name();
        g_main_thread_hub.push_message([host, this]() {
            on_instrument_added.broadcast(host);
        });
    });
}

void plugin_host_manager::process(uint32_t in_frames) const {
    tf::Executor executor;
    tf::Taskflow taskflow;
    for (auto host : plugin_hosts_) {
        taskflow.emplace([host, in_frames] {
            host->process(in_frames);
        });
    }
    executor.run(taskflow).wait();
}

void plugin_host_manager::on_mixer_track_removed(mixer_track* track) {
    for (auto host : plugin_hosts_) {
        host->channel->remove_track(track);
    }
}
