#include "vst2_plugin_host.h"

#include "audio/device/audio_device_manager.h"
#include "audio/mixer/channel_interface.h"

std::map<std::string, std::weak_ptr<dynamic_library>> vst2_library_map;
VstTimeInfo vst2_plugin_host::vst_time_info{};

std::map<std::string, bool> can_do_map =
{
    {"sendVstEvents", true},
    {"sendVstMidiEvent", true},
    {"sendVstTimeInfo", true},
    {"receiveVstEvents", false},
    {"receiveVstMidiEvent", false},
    {"receiveVstTimeInfo", false},
    {"offline", false},
    {"reportConnectionChanges", false},
    {"acceptIOChanges", true},
    {"sizeWindow", true},
    {"asyncProcessing", true},
    {"supplyIdle", true},
    {"supportShell", false},
    {"openFileSelector", false},
    {"editFile", false},
    {"closeFileSelector", false},
    {"NIMKPIVendorSpecificCallbacks", false},
};

std::shared_ptr<dynamic_library> get_vst2_library(const std::string& path) {
    if (vst2_library_map.contains(path) && !vst2_library_map[path].expired()) {
        return vst2_library_map[path].lock();
    }
    auto library = std::make_shared<dynamic_library>(path);
    vst2_library_map[path] = library;
    return library;
}

bool vst_host_can_do(const char* can_do)
{
    if (can_do_map.contains(can_do))
        return can_do_map[can_do];
    spdlog::warn("Unknown can do: {}", can_do);
    return false;
}

VstIntPtr vst_master_callback(AEffect* effect, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
    switch (opcode)
    {
        case audioMasterAutomate:
            return 1;
        case audioMasterVersion:
            return kVstVersion; // 返回插件版本号
        case audioMasterWantMidi:
            return 1;
        case audioMasterGetSampleRate:
            return g_audio_device_manager.get_sample_rate(); // 返回采样率
        case audioMasterGetCurrentProcessLevel:
            return kVstProcessLevelRealtime; // 返回当前处理级别
        case audioMasterGetBlockSize:
            return g_audio_device_manager.get_buffer_size(); // 返回块大小
        case audioMasterSizeWindow:
        {
            // 设置插件窗口大小
            // g_window_manager.resize_plugin_host_window((plugin_host*)effect->user, index, value);
            return 1;
        }
        case audioMasterGetTime:
            return (VstIntPtr)&vst2_plugin_host::vst_time_info;
        case audioMasterIdle:
            // UE_LOG(LogTemp, Log, TEXT("Plugin Idle"));
        return 1;
        case audioMasterNeedIdle:
            // UE_LOG(LogTemp, Log, TEXT("Plugin Need Idle"));
        return 1;
        case audioMasterGetVendorString:
        {
            const char* Ptr1 = (const char*)ptr;
            Ptr1 = "Arona";
            return 1;
        }
        case audioMasterGetProductString:
        {
            const char* Ptr1 = (const char*)ptr;
            Ptr1 = "Arona";
            return 1;
        }
        case audioMasterGetVendorVersion:
            return 1000;
        case audioMasterCanDo:
        {
            const char* Ptr1 = (const char*)ptr;
            return vst_host_can_do(Ptr1) ? 1 : 0;
        }
        case audioMasterUpdateDisplay:
            return 1;
        case audioMasterBeginEdit:
        {
            // UE_LOG(LogTemp, Log, TEXT("Want Begin Edit"));
            return 1;
        }
        case audioMasterEndEdit:
        {
            // UE_LOG(LogTemp, Log, TEXT("Want End Edit"));
            return 1;
        }
        case audioMasterVendorSpecific:
            return 0;
        case audioMasterProcessEvents:
            // TODO
        {
            VstEvents* Events = (VstEvents*)ptr;
            Events->events[0]->type;
            // FVST2PluginHost* Host = static_cast<FVST2PluginHost*>(Effect->user);

        }
        return 1;
        default: break;
            // ensureMsgf(0, TEXT("No Implement OpCode: %d"), OpCode);
    }

    return 1;
}

typedef AEffect*(*vst_plugin_entry_proc)(audioMasterCallback AudioMaster);

vst2_plugin_host::vst2_plugin_host() {
    effect_ = nullptr;
    library_ = nullptr;
}

vst2_plugin_host::~vst2_plugin_host() {
    spdlog::info("vst2 plugin {} destroyed", name);
    try_close_editor();
    if (effect_) {
        dispatch(effClose);
    }
}

bool vst2_plugin_host::load_plugin(const char* path) {
    // load vst
    library_ = get_vst2_library(path);
    // get main function
    const auto entry_proc = (vst_plugin_entry_proc)(library_->get_function("VSTPluginMain"));
    if (!entry_proc) {
        spdlog::error("Failed to get entry point from {}", path);
        return false;
    }
    // create effect
    effect_ = entry_proc(&vst_master_callback);
    if (!effect_) {
        return false;
    }
    if (effect_->magic != kEffectMagic)
        throw std::runtime_error("Not a VST2 plugin");
    effect_->user = this;

    name = load_name();
    vendor = load_vendor();

    dispatch(effOpen);
    return true;
}

void vst2_plugin_host::set_enabled(bool enabled) {
    enabled_ = enabled;
    dispatch(effMainsChanged, 0, enabled);
}

bool vst2_plugin_host::is_enabled() const {
    return enabled_;
}

void vst2_plugin_host::on_update_sample_rate(double sample_rate) {
    dispatch(effSetSampleRate, 0, 0, nullptr, sample_rate);
}

void vst2_plugin_host::on_update_buffer_size(int buffer_size) {
    dispatch(effSetBlockSize, 0, buffer_size);
}

uint32_t vst2_plugin_host::get_input_channels() const {
    return effect_->numInputs;
}

uint32_t vst2_plugin_host::get_output_channels() const {
    return effect_->numOutputs;
}

void vst2_plugin_host::update_channel_node_name() {
    for (int i = 0; i < effect_->numInputs; ++i)
    {
        VstPinProperties pin_properties{};
        if (dispatch(effGetInputProperties, i, 0, &pin_properties) == 1)
        {
            channel->set_input_channel_node_name(i / 2, i, pin_properties.label);
        }
    }

    for (int i = 0; i < effect_->numOutputs; ++i)
    {
        VstPinProperties pin_properties{};
        if (dispatch(effGetOutputProperties, i, 0, &pin_properties) == 1)
        {
            channel->set_output_channel_node_name(i / 2, i, pin_properties.label);
        }
    }
}

void vst2_plugin_host::process(uint32_t frame_num) {
    // TODO send midi


    effect_->processReplacing(effect_, channel->get_input_headers(), channel->get_output_headers(), frame_num);
}

void vst2_plugin_host::idle_editor() {
    plugin_host::idle_editor();
    dispatch(effEditIdle);
}

bool vst2_plugin_host::has_editor() const {
    return effect_->flags & effFlagsHasEditor;
}

// ImVec2 vst2_plugin_host::get_editor_size() const {
//     ERect* EditorRect = nullptr;
//     dispatch(effEditGetRect, 0, 0, &EditorRect);
//     if (EditorRect)
//         return ImVec2(EditorRect->right - EditorRect->left, EditorRect->bottom - EditorRect->top);
//     return ImVec2(0, 0);
// }

std::string vst2_plugin_host::load_name() const {
    char buffer[256];
    dispatch(effGetEffectName, 0, 0, buffer);
    return buffer;
}

std::string vst2_plugin_host::load_vendor() const {
    char buffer[256];
    dispatch(effGetVendorString, 0, 0, buffer);
    return buffer;
}

void vst2_plugin_host::open_editor() {
    if (!has_editor())
        return;
    // void* window_handle = glfwGetWindowHandle(window);

    // dispatch(effEditOpen, 0, 0, window_handle);
}

void vst2_plugin_host::close_editor() {
    if (!has_editor())
        return;
    dispatch(effEditClose);
}

VstIntPtr vst2_plugin_host::dispatch(VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt) const {
    return effect_->dispatcher(effect_, opcode, index, value, ptr, opt);
}
